/* keyframebind.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */
#include <stdio.h>

#include "keyframe.h"
#include "keyframelist.h"
#include "keyframebind.h"

KeyFrameBind *
createKeyFrameBind( const GLuint id )
{
  KeyFrameBind *ret = (KeyFrameBind *)malloc( sizeof( KeyFrameBind ) );
  ret->ref = 1;
  ret->id = id;
  ret->bone = NULL;
  ret->rotate1 = NULL;
  ret->rotate2 = NULL;
  ret->next = NULL;
  return ret;
}

void 
destroyKeyFrameBind( KeyFrameBind *k )
{
  if( k!=NULL ){
    if( --k->ref<=0 ){
      destroyModelBone( k->bone );
      destroyKeyFrameLists( k->rotate1 );
      destroyKeyFrameLists( k->rotate2 );
      free( k );
    }
  }
}

void 
destroyKeyFrameBinds( KeyFrameBind *begin )
{
  KeyFrameBind *target = begin, *reserve = NULL;
  while( target!=NULL ){
    reserve = target;
    target = target->next;
    destroyKeyFrameBind( reserve );
  }
}

void 
setModelBoneKeyFrameBind( KeyFrameBind *k, ModelBone *m )
{
  if( k!=NULL ){
    if( k->bone!=NULL ){
      destroyModelBone( k->bone );
    }
    k->bone = m;
    if( m!=NULL ){
      m->ref++;
    }
  }
}

ModelBone *
getModelBoneKeyFrameBind( KeyFrameBind *k )
{
  ModelBone *ret = NULL;
  if( k!=NULL ){
    if( k->bone!=NULL ){
      ret = k->bone;
      ret->ref++;
    }
  }
  return ret;
}

/* link */
void 
linkKeyFrameBind( KeyFrameBind *target, KeyFrameBind *branch )
{
  KeyFrameBind *last;
  
  if( target!=NULL ){
    last = target;
    while( last->next!=NULL ){
      last = last->next;
    }
    last->next = branch;

    if( branch!=NULL ){
      branch->ref++;
    }
  }
}

KeyFrameBind *
nextKeyFrameBind( KeyFrameBind *target )
{
  KeyFrameBind *ret = NULL;
  if( target!=NULL ){
    ret = target->next;
  }
  return ret;
}

/* add new animation */
void 
addKeyFrameKeyFrameBind( KeyFrameBind *target, KeyFrame *begin, const double current_time )
{
  KeyFrame *k = begin, *state_rotate1, *state_rotate2;
  KeyFrameList *rotate1, *rotate2;
  KeyValueAbstract *value_rotate1, *value_rotate2;

  /* create key list */
  rotate1 = createKeyFrameList();
  rotate2 = createKeyFrameList();

  /* set state key frame */
  state_rotate1 = createKeyFrame( current_time );
  state_rotate2 = createKeyFrame( current_time );
  value_rotate1 = createKeyValue( KEY_VALUE_TYPE_ROTATE1, target->id );
  value_rotate2 = createKeyValue( KEY_VALUE_TYPE_ROTATE2, target->id );
  value_rotate1->rotate1.degree = target->bone->rotate1;
  value_rotate2->rotate2.degree = target->bone->rotate2;
  setKeyValueToKeyFrame( state_rotate1, value_rotate1 );
  setKeyValueToKeyFrame( state_rotate2, value_rotate2 );
  destroyKeyValue( value_rotate1 );
  destroyKeyValue( value_rotate2 );

  while( k!=NULL ){
    KeyValueAbstract *value = getKeyValueFromKeyFrame( k ), *tmp;
    while( value!=NULL ){
      /* 
      fprintf( stderr, "(Timing = %f ) / VALUE ID = %d / TARGET ID = %d\n", 
        getTimingKeyFrame( k ), value->common.id, target->id );
      */
      if( value->common.id==target->id ){
        KeyFrame *add = createKeyFrame( getTimingKeyFrame( k ) + current_time );

        tmp = createKeyValue( KEY_VALUE_TYPE_NONE, target->id );
        copyKeyValue( tmp, value );
        setKeyValueToKeyFrame( add, tmp );

        switch( value->common.type ){
        case KEY_VALUE_TYPE_ROTATE1:
          linkKeyFrame( state_rotate1, add );
          break;
        case KEY_VALUE_TYPE_ROTATE2:
          linkKeyFrame( state_rotate2, add );
          break;
        }
        /*
        fprintf( stderr, "set key %d\n", value->common.type );
        */
        destroyKeyValue( tmp );
        destroyKeyFrame( add );
      }
      value = nextKeyValue( value );
    }
    destroyKeyValue( value );
    k = nextKeyFrame( k );
  }

  /* KEY_VALUE_TYPE_ROTATE1 */
  if( target->rotate1!=NULL ){
    linkKeyFrameList( target->rotate1, rotate1 );
  } else {
    target->rotate1 = rotate1;
    rotate1->ref++;
  }
  /* KEY_VALUE_TYPE_ROTATE2 */
  if( target->rotate2!=NULL ){
    linkKeyFrameList( target->rotate2, rotate2 );
  } else {
    target->rotate2 = rotate2;
    rotate2->ref++;
  }

  setKeyFrameToKeyFrameList( rotate1, state_rotate1 );
  setKeyFrameToKeyFrameList( rotate2, state_rotate2 );
  destroyKeyFrame( state_rotate1 );
  destroyKeyFrame( state_rotate2 );
  destroyKeyFrameList( rotate1 );
  destroyKeyFrameList( rotate2 );
}

/* execution utility */
static KeyValueAbstract *
applyKeyValue( KeyFrameBind *k, KeyFrameList *list, const int type, 
               const double current_time, int *layer )
{
  double timing, timing_prev;
  KeyFrame *frame, *frame_prev;

  KeyValueAbstract *value = NULL;
  *layer = 0;

  while( list!=NULL ){
    frame = list->key_frame;
    frame_prev = NULL;

    while( frame!=NULL ){
      timing=getTimingKeyFrame( frame );
      if( timing<current_time ){
        frame_prev = frame;
        timing_prev = timing;
      } else if( frame_prev!=NULL ) {
        KeyValueAbstract *add = createKeyValue( type, k->id );

        /* Interpolation */
        frame->list->common.linearInterpolation( (void *)add, (void *)frame_prev->list, (void *)frame->list, 
          (current_time - timing_prev)/(timing - timing_prev) );
        
        if( value!=NULL ){
          linkKeyValue( value, add );
        } else {
          value = add;
          value->common.ref++;
        }

        /*
        fprintf( stderr, "[type=%d/(%f:%f)(%f:%f)(%f:%f)]\n", type, 
          timing_prev, frame_prev->list->rotate1.degree, 
          current_time, add->rotate1.degree, 
          timing, frame->list->rotate1.degree );
        */
        destroyKeyValue( add );
        (*layer)++;
        break;
      }
      frame = nextKeyFrame( frame );
    }
    list = list->next;
  }
  return value;
}
void 
applyKeyFrameBind( KeyFrameBind *k, const double current_time )
{
  int layer;
  KeyValueAbstract *value;

  /* apply ROTATE1 */
  value = applyKeyValue( k, k->rotate1, KEY_VALUE_TYPE_ROTATE1, current_time, &layer );
  if( layer>0 && k->bone!=NULL ){
    KeyValueAbstract *v = value;
    ModelBone *bone = k->bone;
    bone->rotate1 = 0.0;
    while( v!=NULL ){
      /* assembling */
      bone->rotate1 += v->rotate1.degree;
      v = nextKeyValue( v );
    }
    bone->rotate1 /= (double)layer;
  }
  destroyKeyValues( value );

  /* apply ROTATE2 */
  value = applyKeyValue( k, k->rotate2, KEY_VALUE_TYPE_ROTATE2, current_time, &layer );
  if( layer>0 && k->bone!=NULL ){
    KeyValueAbstract *v = value;
    ModelBone *bone = k->bone;
    bone->rotate2 = 0.0;
    while( v!=NULL ){
      /* assembling */
      bone->rotate2 += v->rotate2.degree;
      v = nextKeyValue( v );
    }
    bone->rotate2 /= (double)layer;
  }
  destroyKeyValues( value );
}

static KeyFrameList *
expireKeyFrameBind( KeyFrameList *list, const double current_time )
{
  KeyFrameList *first = list, *prev = NULL;
  int check;

  while( list!=NULL ){
    KeyFrame *frame = list->key_frame;

    check = 0;
    while( frame!=NULL ){
      if( getTimingKeyFrame( frame )>current_time ){
        check++;
        break;
      }
      frame = nextKeyFrame( frame );
    }
    if( check==0 ){
      /* expire this list */
      KeyFrameList *expire = list;
      if( prev==NULL ){
        first = list->next;
        list = first;
      } else {
        prev->next = list->next;
        list = list->next;
      }
      destroyKeyFrameList( expire );
    } else {
      prev = list;
      list = list->next;
    }
  }
  return first;
}
void 
expireOldKeyFrameKeyFrameBind( KeyFrameBind *k, const double current_time )
{
  if( k!=NULL ){
    k->rotate1 = expireKeyFrameBind( k->rotate1, current_time );
    k->rotate2 = expireKeyFrameBind( k->rotate2, current_time );
  }
}
