/* modelfile.c	-- Yasuhiro Yamazaki Mon Dec 24 2001
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>

#include <GL/gl.h>

#include "vector3d.h"
#include "modelbone.h"
#include "modelfile.h"
#include "utility.h"

/* create */
static ModelFile *
createModelFile( void )
{
  ModelFile *m = (ModelFile *)malloc( sizeof( ModelFile ) );
  m->ref = 1;
  m->filename = NULL;
  m->root = NULL;
  m->list = NULL;
  return m;
}

/* set filename */
static void
setFilenameModelFile( ModelFile *m, const char *filename )
{
  if( m!=NULL ){
    if( m->filename!=NULL ){
      free( m->filename );
    }
    m->filename = (char *)malloc( (strlen( filename )+1)*sizeof(char) );
    strcpy( m->filename, filename );
  }
}

/* create list node */
static ModelBoneList *
createModelBoneList( ModelBone *model )
{
  ModelBoneList *l = (ModelBoneList *)malloc( sizeof( ModelBoneList ) );
  l->model = model;
  l->next = NULL;
  return l;
}

/* add */
static void
addModelBoneListToModelFile( ModelFile *mf, ModelBoneList *mbl )
{
  if( mf!=NULL && mbl!=NULL ){
    mbl->next = mf->list;
    mf->list = mbl;
  }
}

/* load from stream */
static ModelBone *state = NULL;
static ModelBone *
loadModelFileFromStream( ModelFile *m, FILE *read, ModelBone *parent )
{
  char tmp[256];
  ModelBone *brother = NULL;
  state = NULL;
  
  while( !feof( read ) ){
    if( fscanf( read, "%s", tmp )!=EOF ){
      stringToLower( tmp );

      if( *tmp=='#' ){
	/* comment by '\n' */
	while( !feof( read ) && fgetc( read )!='\n' );
	
      } else if( strcmp( tmp, "loadfile" )==0 ){
	/* load state model */
        ModelBoneList *add = NULL;
	fscanf( read, "%s", tmp );
        if( state!=NULL ){
          destroyModelBone( state );
        }
	state = createModelBone( tmp );
        state->ref++;
	if( brother!=NULL ){
	  linkAsBrotherModelBone( brother, state );
	} else {
	  brother = state;
          brother->ref++;
	}
	addModelBoneListToModelFile( m, (add=createModelBoneList( state )) );
        destroyModelBone( state );

      } else if( strcmp( tmp, "id" )==0 ){
	/* set ID Number to state model */
	GLuint id = 0;
	fscanf( read, "%d", &id );
	setIDNumberModelBone( state, id );
	
      } else if( strcmp( tmp, "scale" )==0 ){
	/* set scale to state model */
	Vector3D scale;
	fscanf( read, "%f%f%f", &scale.x, &scale.y, &scale.z );
	setScaleModelBone( state, scale );
	
      } else if( strcmp( tmp, "translate" )==0 ){
	/* set translate to state model */
	Vector3D translate;
	fscanf( read, "%f%f%f", &translate.x, &translate.y, &translate.z );
	setTranslateModelBone( state, translate );
	
      } else if( strcmp( tmp, "center" )==0 ){
	/* set center to state model */
	Vector3D center;
	fscanf( read, "%f%f%f", &center.x, &center.y, &center.z );
	setCenterModelBone( state, center );
	
      } else if( strcmp( tmp, "rotate1" )==0 ){
	/* set rotate1 to state model */
	double rotate;
	Vector3D axis;
	fscanf( read, "%lf%f%f%f", &rotate, &axis.x, &axis.y, &axis.z );
	setRotate1ModelBone( state, rotate );
	setAxis1ModelBone( state, axis );
	
      } else if( strcmp( tmp, "rotate2" )==0 ){
	/* set rotate2 to state model */
	double rotate;
	Vector3D axis;
	fscanf( read, "%lf%f%f%f", &rotate, &axis.x, &axis.y, &axis.z );
	setRotate2ModelBone( state, rotate );
	setAxis2ModelBone( state, axis );
	
      } else if( strcmp( tmp, "{" )==0 ){
	/* stack on */
	ModelBone *parent = state;
	linkAsChildModelBone( parent, loadModelFileFromStream( m, read, parent ) );
	
      } else if( strcmp( tmp, "}" )==0 ){
	/* stack off */
	return brother;
	
      } else {
	/* ERROR */
	fprintf( stderr, "Syntax error \"%s\" in line.\n", tmp );
      }
    }
  }
  
  return brother;
}

/* load */
ModelFile *
loadModelFile( const char *filename )
{
  FILE *read;
  ModelFile *m = createModelFile();
  setFilenameModelFile( m, filename );
  
  /* loading */
  read = fopen( filename, "r" );
  if( read!=NULL ){
    m->root = loadModelFileFromStream( m, read, NULL );
    if( state!=NULL ){
      destroyModelBone( state );
      state = NULL;
    }
    fclose( read );
  }
  
  return m;
}

/* destroy model list and model */
static void
destroyModelBoneList( ModelBoneList *list )
{
  if( list!=NULL ){
    destroyModelBoneList( list->next );
    destroyModelBone( list->model );
    free( list );
  }
}

/* destroy */
void 
destroyModelFile( ModelFile *m )
{
  if( m!=NULL ){
    if( --m->ref<=0 ){
      if( m->filename!=NULL ){
        free( m->filename );
        m->filename = NULL;
      }
      destroyModelBoneList( m->list );
      free( m );
    }
  }
}

/* put */
void 
putModelFile( ModelFile *m )
{
  if( m!=NULL ){
    putLinkedModelBone( m->root );
  }
}

/* get */
ModelBone *
getModelBoneFromModelFile( ModelFile *m )
{
  if( m!=NULL ){
    return m->root;
  }
  return NULL;
}
